//
//  FilterWordsUseCase.swift
//  ReviewsParserKit
//
//  Created by Javier Cicchelli on 17/03/2024.
//  Copyright © 2024 Röck+Cöde VoF. All rights reserved.
//

import Foundation

public struct FilterWordsUseCase {
    
    // MARK: Type aliases
    public typealias Input = String
    public typealias Output = [Tuple]
    
    // MARK: Initialisers
    public init() {}
    
    // MARK: Functions
    public func callAsFunction(_ input: Input) throws -> Output {
        let regularExpression = try NSRegularExpression(
            pattern: .Pattern.wordsLongerThan4Characters,
            options: .caseInsensitive
        )
        
        let matches = regularExpression.matches(
            in: input,
            options: [],
            range: .init(input.startIndex..., in: input)
        )
        
        guard !matches.isEmpty else { return [] }
        
        let wordsAll = matches.compactMap {
            Range($0.range, in: input).map { String(input[$0]) }
        }
        
        let wordsUnique = wordsAll
            .reduce(into: [String]()) { partialResult, word in
                guard partialResult
                    .filter({ compareWords(word, $0) })
                    .isEmpty
                else { return }
                
                partialResult.append(
                    word.folding(
                        options: [
                            .caseInsensitive,
                            .diacriticInsensitive
                        ],
                        locale: .current
                    )
                    .capitalized
                )
            }
        
        return Dictionary(grouping: wordsUnique) { word in
            wordsAll
                .filter { compareWords(word, $0) }
                .count
        }
        .flatMap { (key, values) -> Output in
            values.map {
                .init(word: $0, count: key)
            }
        }
        .sorted {
            guard $0.count != $1.count else {
                return $0.word < $1.word
            }
            
            return $0.count > $1.count
        }
    }
    
}

// MARK: - Helpers
private extension FilterWordsUseCase {
    
    // MARK: Functions
    func compareWords(
        _ lword: String,
        _ rword: String
    ) -> Bool {
        lword.compare(rword, options: [
            .caseInsensitive,
            .diacriticInsensitive
        ]) == .orderedSame
    }
    
}

// MARK: - Structs
extension FilterWordsUseCase {
    public struct Tuple: Equatable {
        
        // MARK: Constants
        let word: String
        let count: Int
        
        // MARK: Initialisers
        public init(
            word: String,
            count: Int
        ) {
            self.word = word
            self.count = count
        }

    }
}

// MARK: - String+Constants
private extension String {
    enum Pattern {
        static let wordsLongerThan4Characters = "\\w{4,}"
    }
}
